/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package es.rickyepoderi.apacheds;

import java.io.File;
import java.net.URL;
import java.util.List;
import java.util.Map;
import java.util.TreeSet;
import java.util.UUID;
import java.util.regex.Pattern;
import javax.naming.InvalidNameException;
import net.sf.ehcache.CacheManager;
import net.sf.ehcache.config.CacheConfiguration;
import net.sf.ehcache.config.Configuration;
import org.apache.commons.io.IOUtils;
import org.apache.directory.api.ldap.model.constants.SchemaConstants;
import org.apache.directory.api.ldap.model.entry.DefaultEntry;
import org.apache.directory.api.ldap.model.entry.Entry;
import org.apache.directory.api.ldap.model.exception.LdapEntryAlreadyExistsException;
import org.apache.directory.api.ldap.model.ldif.LdifEntry;
import org.apache.directory.api.ldap.model.ldif.LdifReader;
import org.apache.directory.api.ldap.model.schema.LdapComparator;
import org.apache.directory.api.ldap.model.schema.SchemaManager;
import org.apache.directory.api.ldap.model.schema.comparators.NormalizingComparator;
import org.apache.directory.api.ldap.model.schema.registries.ComparatorRegistry;
import org.apache.directory.api.ldap.model.schema.registries.SchemaLoader;
import org.apache.directory.api.ldap.schema.extractor.impl.DefaultSchemaLdifExtractor;
import org.apache.directory.api.ldap.schema.extractor.impl.ResourceMap;
import org.apache.directory.api.ldap.schema.loader.JarLdifSchemaLoader;
import org.apache.directory.api.ldap.schema.manager.impl.DefaultSchemaManager;
import org.apache.directory.api.util.exception.Exceptions;
import org.apache.directory.server.constants.ServerDNConstants;
import org.apache.directory.server.core.DefaultDirectoryService;
import org.apache.directory.server.core.api.CacheService;
import org.apache.directory.server.core.api.DirectoryService;
import org.apache.directory.server.core.api.InstanceLayout;
import org.apache.directory.server.core.api.interceptor.Interceptor;
import org.apache.directory.server.core.api.interceptor.context.AddOperationContext;
import org.apache.directory.server.core.api.partition.Partition;
import org.apache.directory.server.core.api.schema.SchemaPartition;
import org.apache.directory.server.core.factory.AvlPartitionFactory;
import org.apache.directory.server.core.normalization.NormalizationInterceptor;
import org.apache.directory.server.core.partition.ldif.AbstractLdifPartition;
import org.apache.directory.server.i18n.I18n;
import org.apache.directory.server.ldap.LdapServer;
import org.apache.directory.server.protocol.shared.transport.TcpTransport;

/**
 *
 * @author ricky
 */
public class Server {

    static class InMemorySchemaPartition extends AbstractLdifPartition {

        public InMemorySchemaPartition(SchemaManager schemaManager) {
            super(schemaManager);
        }

        @Override
        protected void doInit() throws InvalidNameException, Exception {
            if (initialized) {
                return;
            }
            suffixDn.apply(schemaManager);
            super.doInit();

            // load schema
            final Map<String, Boolean> resMap = ResourceMap.getResources(Pattern.compile("schema[/\\Q\\\\E]ou=schema.*"));
            for (String resourcePath : new TreeSet<>(resMap.keySet())) {
                if (resourcePath.endsWith(".ldif")) {
                    URL resource = DefaultSchemaLdifExtractor.getUniqueResource(resourcePath, "Schema LDIF file");
                    LdifEntry ldifEntry;
                    try (LdifReader reader = new LdifReader(resource.openStream())) {
                        ldifEntry = reader.next();
                    }
                    Entry entry = new DefaultEntry(schemaManager, ldifEntry.getEntry());

                    // add mandatory attributes
                    if (entry.get(SchemaConstants.ENTRY_CSN_AT) == null) {
                        entry.add(SchemaConstants.ENTRY_CSN_AT, AbstractLdifPartition.defaultCSNFactory.newInstance().toString());
                    }
                    if (entry.get(SchemaConstants.ENTRY_UUID_AT) == null) {
                        entry.add(SchemaConstants.ENTRY_UUID_AT, UUID.randomUUID().toString());
                    }
                    AddOperationContext addContext = new AddOperationContext(null, entry);
                    super.add(addContext);
                }
            }
        }

    }
    
    private static void importLdifContent(DirectoryService directoryService, String ldifContent) throws Exception {
        try (LdifReader ldifReader = new LdifReader(IOUtils.toInputStream(ldifContent))) {
            for (LdifEntry ldifEntry : ldifReader) {
                try {
                    System.err.println(directoryService.getAdminSession());
                    directoryService.getAdminSession().add(new DefaultEntry(directoryService.getSchemaManager(), ldifEntry.getEntry()));
                } catch (LdapEntryAlreadyExistsException ignore) {
                    System.err.println("Entry " + ldifEntry.getDn() + " already exists. Ignoring");
                }
            }
        }
    }

    private final DirectoryService directoryService;
    
    public Server(int port) throws Exception {
        directoryService = new DefaultDirectoryService();
        directoryService.setShutdownHookEnabled(true);
        directoryService.setAllowAnonymousAccess(true);
        InstanceLayout instanceLayout = new InstanceLayout(System.getProperty("java.io.tmpdir"));
        directoryService.setInstanceLayout(instanceLayout);

        Configuration ehCacheConfig = new Configuration();
        ehCacheConfig.setName("system");
        CacheConfiguration defaultCache = new CacheConfiguration("system-default", 1).eternal(false).timeToIdleSeconds(30)
                .timeToLiveSeconds(30).overflowToDisk(false);
        ehCacheConfig.addDefaultCache(defaultCache);
        CacheService cacheService = new CacheService(new CacheManager(ehCacheConfig));
        directoryService.setCacheService(cacheService);

        SchemaLoader loader = new JarLdifSchemaLoader();
        SchemaManager schemaManager = new DefaultSchemaManager(loader);
        schemaManager.loadAllEnabled();
        ComparatorRegistry comparatorRegistry = schemaManager.getComparatorRegistry();
        for (LdapComparator<?> comparator : comparatorRegistry) {
            if (comparator instanceof NormalizingComparator) {
                ((NormalizingComparator) comparator).setOnServer();
            }
        }
        directoryService.setSchemaManager(schemaManager);
        InMemorySchemaPartition inMemorySchemaPartition = new InMemorySchemaPartition(schemaManager);
        SchemaPartition schemaPartition = new SchemaPartition(schemaManager);
        schemaPartition.setWrappedPartition(inMemorySchemaPartition);
        directoryService.setSchemaPartition(schemaPartition);
        List<Throwable> errors = schemaManager.getErrors();
        if (!errors.isEmpty()) {
            throw new Exception(I18n.err(I18n.ERR_317, Exceptions.printErrors(errors)));
        }

        Partition systemPartition = new AvlPartitionFactory().createPartition(
                directoryService.getSchemaManager(),
                directoryService.getDnFactory(),
                "system",
                ServerDNConstants.SYSTEM_DN,
                500,
                new File(directoryService.getInstanceLayout().getPartitionsDirectory(), "system"));
        systemPartition.setSchemaManager(directoryService.getSchemaManager());
        systemPartition.setCacheService(directoryService.getCacheService());
        directoryService.setSystemPartition(systemPartition);
        
        Partition samplePartition = new AvlPartitionFactory().createPartition(
                directoryService.getSchemaManager(),
                directoryService.getDnFactory(),
                "sample",
                "dc=sample,dc=com",
                500,
                new File(directoryService.getInstanceLayout().getPartitionsDirectory(), "sample"));
        samplePartition.setSchemaManager(directoryService.getSchemaManager());
        samplePartition.setCacheService(directoryService.getCacheService());
        directoryService.addPartition(samplePartition);

        List<Interceptor> interceptors = directoryService.getInterceptors();

        // Find Normalization interceptor in chain
        int insertionPosition = -1;
        for (int pos = 0; pos < interceptors.size(); ++pos) {
            Interceptor interceptor = interceptors.get(pos);
            if (interceptor instanceof NormalizationInterceptor) {
                insertionPosition = pos;
            }
        }

        // insert our new interceptor just behind
        interceptors.add(insertionPosition + 1, new RangedAttributeInterceptor("uniqueMember", 5));
        directoryService.setInterceptors(interceptors);

        LdapServer ldapServer = new LdapServer();
        ldapServer.setDirectoryService(directoryService);

        TcpTransport ldapTransport = new TcpTransport(port);
        ldapServer.setTransports(ldapTransport);

        directoryService.startup();
        ldapServer.start();
        
        String entryLdif
                = "dn: dc=sample,dc=com\n"
                + "dc: sample\n"
                + "objectClass: top\n"
                + "objectClass: domain\n"
                + "\n"
                + "dn: ou=People,dc=sample,dc=com\n"
                + "ou: People\n"
                + "objectClass: top\n"
                + "objectClass: organizationalUnit\n"
                + "\n"
                + "dn: ou=Groups,dc=sample,dc=com\n"
                + "ou: Groups\n"
                + "objectClass: top\n"
                + "objectClass: organizationalUnit\n"
                + "\n"
                + "dn: uid=user0,ou=People,dc=sample,dc=com\n"
                + "uid: user0\n"
                + "sn: user0\n"
                + "cn: user0\n"
                + "objectClass: top\n"
                + "objectClass: organizationalPerson\n"
                + "objectClass: inetOrgPerson\n"
                + "\n"
                + "dn: uid=user1,ou=People,dc=sample,dc=com\n"
                + "uid: user1\n"
                + "sn: user1\n"
                + "cn: user1\n"
                + "objectClass: top\n"
                + "objectClass: organizationalPerson\n"
                + "objectClass: inetOrgPerson\n"
                + "\n"
                + "dn: uid=user2,ou=People,dc=sample,dc=com\n"
                + "uid: user2\n"
                + "sn: user2\n"
                + "cn: user2\n"
                + "objectClass: top\n"
                + "objectClass: organizationalPerson\n"
                + "objectClass: inetOrgPerson\n"
                + "\n"
                + "dn: uid=user3,ou=People,dc=sample,dc=com\n"
                + "uid: user3\n"
                + "sn: user3\n"
                + "cn: user3\n"
                + "objectClass: top\n"
                + "objectClass: organizationalPerson\n"
                + "objectClass: inetOrgPerson\n"
                + "\n"
                + "dn: uid=user4,ou=People,dc=sample,dc=com\n"
                + "uid: user4\n"
                + "sn: user4\n"
                + "cn: user4\n"
                + "objectClass: top\n"
                + "objectClass: organizationalPerson\n"
                + "objectClass: inetOrgPerson\n"
                + "\n"
                + "dn: uid=user5,ou=People,dc=sample,dc=com\n"
                + "uid: user5\n"
                + "sn: user5\n"
                + "cn: user5\n"
                + "objectClass: top\n"
                + "objectClass: organizationalPerson\n"
                + "objectClass: inetOrgPerson\n"
                + "\n"
                + "dn: uid=user6,ou=People,dc=sample,dc=com\n"
                + "uid: user6\n"
                + "sn: user6\n"
                + "cn: user6\n"
                + "objectClass: top\n"
                + "objectClass: organizationalPerson\n"
                + "objectClass: inetOrgPerson\n"
                + "\n"
                + "dn: uid=user7,ou=People,dc=sample,dc=com\n"
                + "uid: user7\n"
                + "sn: user7\n"
                + "cn: user7\n"
                + "objectClass: top\n"
                + "objectClass: organizationalPerson\n"
                + "objectClass: inetOrgPerson\n"
                + "\n"
                + "dn: uid=user8,ou=People,dc=sample,dc=com\n"
                + "uid: user8\n"
                + "sn: user8\n"
                + "cn: user8\n"
                + "objectClass: top\n"
                + "objectClass: organizationalPerson\n"
                + "objectClass: inetOrgPerson\n"
                + "\n"
                + "dn: uid=user9,ou=People,dc=sample,dc=com\n"
                + "uid: user9\n"
                + "sn: user9\n"
                + "cn: user9\n"
                + "objectClass: top\n"
                + "objectClass: organizationalPerson\n"
                + "objectClass: inetOrgPerson\n"
                + "\n"
                + "dn: cn=group1,ou=Groups,dc=sample,dc=com\n"
                + "cn: group1\n"
                + "objectClass: top\n"
                + "objectClass: groupOfUniqueNames\n"
                + "uniqueMember: uid=user0,ou=People,dc=sample,dc=com\n"
                + "uniqueMember: uid=user1,ou=People,dc=sample,dc=com\n"
                + "uniqueMember: uid=user2,ou=People,dc=sample,dc=com\n"
                + "uniqueMember: uid=user3,ou=People,dc=sample,dc=com\n"
                + "uniqueMember: uid=user4,ou=People,dc=sample,dc=com\n"
                + "uniqueMember: uid=user5,ou=People,dc=sample,dc=com\n"
                + "uniqueMember: uid=user6,ou=People,dc=sample,dc=com\n"
                + "uniqueMember: uid=user7,ou=People,dc=sample,dc=com\n"
                + "uniqueMember: uid=user8,ou=People,dc=sample,dc=com\n"
                + "uniqueMember: uid=user9,ou=People,dc=sample,dc=com\n"
                + "uniqueMember: uid=user10,ou=People,dc=sample,dc=com\n"
                + "\n"
                + "dn: cn=group2,ou=Groups,dc=sample,dc=com\n"
                + "cn: group2\n"
                + "objectClass: top\n"
                + "objectClass: groupOfUniqueNames\n"
                + "uniqueMember: uid=user0,ou=People,dc=sample,dc=com\n"
                + "uniqueMember: uid=user1,ou=People,dc=sample,dc=com\n"
                + "uniqueMember: uid=user2,ou=People,dc=sample,dc=com\n"
                + "\n"
                + "dn: cn=group3,ou=Groups,dc=sample,dc=com\n"
                + "cn: group3\n"
                + "objectClass: top\n"
                + "objectClass: groupOfNames\n"
                + "member: uid=user0,ou=People,dc=sample,dc=com\n"
                + "member: uid=user1,ou=People,dc=sample,dc=com\n"
                + "member: uid=user2,ou=People,dc=sample,dc=com\n"
                + "member: uid=user3,ou=People,dc=sample,dc=com\n"
                + "member: uid=user4,ou=People,dc=sample,dc=com\n"
                + "member: uid=user5,ou=People,dc=sample,dc=com\n"
                + "member: uid=user6,ou=People,dc=sample,dc=com\n"
                + "member: uid=user7,ou=People,dc=sample,dc=com\n"
                + "member: uid=user8,ou=People,dc=sample,dc=com\n"
                + "member: uid=user9,ou=People,dc=sample,dc=com\n"
                + "member: uid=user10,ou=People,dc=sample,dc=com\n"
                + "\n"
        ;
        
        importLdifContent(directoryService, entryLdif);
    }
    
    public void stop() throws Exception {
        directoryService.shutdown();
    }
    
    public static void main(String[] args) throws Exception {
        new Server(10389);
    }
}
